/*-
 * Copyright (c) 2013-2015 Varnish Software AS
 * All rights reserved.
 *
 * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

struct req;
struct vfp_entry;
struct vfp_ctx;

/* Fetch processors --------------------------------------------------*/

#define VFP_DEBUG(ctx, fmt, ...)					\
	do {								\
		if (!DO_DEBUG(DBG_PROCESSORS))				\
			break;						\
		VSLb((ctx)->wrk->vsl, SLT_Debug, "VFP:%s:%d: " fmt,	\
		    __func__, __LINE__, __VA_ARGS__);			\
	} while (0)

enum vfp_status {
	VFP_ERROR = -1,
	VFP_OK = 0,
	VFP_END = 1,
	VFP_NULL = 2,
};

typedef enum vfp_status vfp_init_f(struct vfp_ctx *, struct vfp_entry *);
typedef enum vfp_status
    vfp_pull_f(struct vfp_ctx *, struct vfp_entry *, void *ptr, ssize_t *len);
typedef void vfp_fini_f(struct vfp_ctx *, struct vfp_entry *);

struct vfp {
	const char		*name;
	vfp_init_f		*init;
	vfp_pull_f		*pull;
	vfp_fini_f		*fini;
	const void		*priv1;
};

struct vfp_entry {
	unsigned		magic;
#define VFP_ENTRY_MAGIC		0xbe32a027
	const struct vfp	*vfp;
	void			*priv1;
	intptr_t		priv2;
	enum vfp_status		closed;
	VTAILQ_ENTRY(vfp_entry)	list;
	uint64_t		calls;
	uint64_t		bytes_out;
};

/*--------------------------------------------------------------------
 * VFP filter state
 */

VTAILQ_HEAD(vfp_entry_s, vfp_entry);

struct vfp_ctx {
	unsigned		magic;
#define VFP_CTX_MAGIC		0x61d9d3e5
	int			failed;
	struct http		*req;
	struct http		*resp;
	struct worker		*wrk;
	struct objcore		*oc;
	oaref_t			*p_oc_oaref;
	struct busyobj		*bo;

	struct vfp_entry_s	vfp;
	struct vfp_entry	*vfp_nxt;
	unsigned		obj_flags;
};

enum vfp_status VFP_Suck(struct vfp_ctx *, void *p, ssize_t *lp);
enum vfp_status VFP_Error(struct vfp_ctx *, const char *fmt, ...)
    v_printflike_(2, 3);

/* Deliver processors ------------------------------------------------*/

enum vdp_status {
	VDP_ERROR_FETCH	= -3,	/* Streaming fetch failure */
	VDP_ERROR_IO	= -2,	/* Socket error (ie remote close) */
	VDP_ERROR	= -1,	/* VDP error */
	VDP_OK		= 0,
	VDP_BREAK	= 1,	/* Stop processing (no error) */
};

enum vdp_action {
	VDP_NULL,		/* Input buffer valid after call */
	VDP_FLUSH,		/* Input buffer will be invalidated */
	VDP_LAST,		/* Implies flush, this is the last chunk */
};

typedef int vdp_init_f(struct req *, void **priv);
/*
 * Return value:
 *	negative:	Error - abandon delivery
 *	zero:		OK
 *	positive:	Don't push this VDP anyway
 */

typedef int vdp_fini_f(struct req *, void **priv);
typedef int vdp_bytes_f(struct req *, enum vdp_action, void **priv,
    const void *ptr, ssize_t len);
/* The bytes callback is called for each chunk of data to be processed. To
 * pass data down to the next layer, call VDP_Bytes().
 *
 * The action value determines how to treat the data being processed.
 *
 * If action==VDP_NULL, the data will not be invalidated upon return from
 * the callback function, and the pointers may be kept at any layer.
 *
 * If action==VDP_FLUSH, the data will become invalid upon returning from
 * the callback function. The layer needs to buffer it, or pass it down to
 * the next layer, making sure to call VDP_bytes with action>=VDP_FLUSH.
 *
 * If action==VDP_LAST, this is the very last chunk being processed. This
 * implies VDP_FLUSH. The function needs to call down **once and only
 * once** to the next layer with action==VDP_LAST. If the processing causes
 * the need for multiple chunks to be passed on to the next layer, only
 * the final chunk should have action==VDP_LAST.
 */

typedef void vdp_probe_f(struct req *, void **priv,
    size_t *pstart, ssize_t *pend);
/* The probe callback is called for the first entry only in the VDP chain,
 * if that entry has a probe function defined. It is called with start==0
 * and end==-1, meaning that the object will be passed through the VDP
 * chain in its entirety. The probe function may change the start and end
 * through the pointers in order to limit the parts of the object that are
 * passed through the chain.*/

struct vdp {
	const char		*name;
	vdp_init_f		*init;
	vdp_bytes_f		*bytes;
	vdp_fini_f		*fini;
	vdp_probe_f		*probe;
};

struct vdp_entry {
	unsigned		magic;
#define VDP_ENTRY_MAGIC		0x353eb781
	unsigned		f_seen_last:1;
	unsigned		f_failed_init:1;
	const struct vdp	*vdp;
	void			*priv;
	VTAILQ_ENTRY(vdp_entry)	list;
};

VTAILQ_HEAD(vdp_entry_s, vdp_entry);

struct vdp_ctx {
	unsigned		magic;
#define VDP_CTX_MAGIC		0xee501df7
	struct vdp_entry_s	vdp;
	struct vdp_entry	*nxt;
	int			retval;
};

int VDP_bytes(struct req *, enum vdp_action act, const void *ptr, ssize_t len);
int VDP_Push(struct req *, const struct vdp *, void *priv);
void VDP_probe(struct req *, size_t *pstart, ssize_t *pend);
void VDP_snapshot(struct req *, struct vdp_entry_s *snap);
int VDP_restore(struct req *, struct vdp_entry_s *snap);

/*
 * VCP API
 *
 * Please contact support@varnish-software.com before using.
 * Subject to change, make sure to use "$ABI strict" in your VMOD.
 *
 *  [BACKEND]
 *      |
 *      v
 *   (synthbackend/file)
 *  VCP_VFP_BACKEND
 *  VCP_VFP_DECODE
 *   (gunzip/unbr)
 *  VCP_VFP_TRANSFORM
 *   (xbody)
 *  VCP_VFP_TEXT
 *   (WAF)
 *   (edgestash/ESI)
 *   (gzip/br)
 *  VCP_VFP_ENCODE
 *  VCP_VFP_STORAGE
 *   (crypto)
 *     |
 *     v
 *  [STORAGE]
 *
 */

enum vcp_vfp_pos {
	VCP_VFP_BACKEND,	/* Raw backend body */
	VCP_VFP_DECODE,		/* Raw backend body, content decoding step */
	VCP_VFP_TEXT,		/* Decoded body */
	VCP_VFP_ENCODE,		/* Body encoding step */
	VCP_VFP_STORAGE,	/* Body before entering storage */
	VCP_VFP_TRANSFORM,	/* Decoded body before xbody */
};

enum vfp_status VCP_VFP_Push(VRT_CTX, enum vcp_vfp_pos pos, const struct vfp *,
    void *priv);
