/*-
 * Copyright (c) 2024 Varnish Software AS
 * All rights reserved.
 *
 * Author: Stephane Cance <stephane.cance@varnish-software.com>
 *
 */
#ifdef KVSTORE_JRNL_H_INCLUDED
#  error "kvstore/journal.h included multiple times"
#endif /* KVSTORE_JRNL_H_INCLUDED */
#define KVSTORE_JRNL_H_INCLUDED

/*
 * When the journal is opened, this drives how operations are
 * synchronized with the journal consumer.
 */
enum kvstore_jrnl_mode {
	/* wait for consumer to report full completion */
	KVSTORE_JRNL_SYNC,
	/* wait for consumer to report first attempt result */
	KVSTORE_JRNL_TRY_SYNC,
	/* never wait from progress by the consumer */
	KVSTORE_JRNL_ASYNC,
};

/* as a type: provide a value,	    as an action: invalid */
#define KVSTORE_JRNL_GET                        (1u<<0)
/* as a type: increment & set TTL,  as an action: same as type */
#define KVSTORE_JRNL_INC                        (1u<<1)
/* as a type: set value & TTL,	    as an action: same as type */
#define KVSTORE_JRNL_SET                        (1u<<2)
/* as a type: delete value,	    as an action: same as type */
#define KVSTORE_JRNL_DEL                        (1u<<3)

/* as a type: invalid,		    as an action: operation will be retried */
#define KVSTORE_JRNL_WILL_RETRY			(1u<<14)

/* record valid actions */
#define KVSTORE_JRNL_ENTRY_ACTIONS	(KVSTORE_JRNL_SET | KVSTORE_JRNL_DEL)

struct kvstore_jrnl_rec {
	/* populated by kvstore_jrnl_read(), must not change */
	unsigned			magic;
#define KVSTORE_JRNL_REC_MAGIC			0xaa64c782
	uint32_t			type;
	uint64_t			recid;
	const void			*key;
	size_t				klen;
	const void			*field;
	size_t				fieldlen;
	struct kvstore_jrnl_ref		*ref; /* opaque */
	/* populated by kvstore_jrnl_read() can be changed by the user
	 * before kvstore_jrnl_done() */
	vtim_dur			ttl;
	const char			*v;
	size_t				vsz;
	uint32_t			actions;
};

/*
 * Start using journaling of entries on the kvstore allowing interraction
 * with a log consumer which can both consume the log and set entry values
 * during consumption of the log.
 * The grace value indicates how long in-memory values are to be trusted
 * without triggering events in the journal.
 * The keep value indicates how long negative entries can be kept in memory
 * after the end of the grace time.
 */
void kvstore_jrnl_open(struct kvstore *kv, vtim_dur grace, vtim_dur keep,
    enum kvstore_jrnl_mode m_read, enum kvstore_jrnl_mode m_write,
    enum kvstore_jrnl_mode m_del);

typedef void kvstore_jrnl_wait_sync_f(void *userdata);

/*
 * Set the callback to invoke whenever an operation on the kvstore
 * is blocked waiting for progress from the log consumer.
 * This callback is invoked from the thread waiting for progress.
 */
void kvstore_jrnl_set_wait_sync_cb(struct kvstore *kv,
    kvstore_jrnl_wait_sync_f *cb, void *userdata);

/*
 * Read records from the journal, id has the same semantics as in
 * kvstore_expiration_check() to either save up position or random access.
 * To avoid making allocations, this uses a buffer, the larger the buffer the
 * larger the list of records can be.
 * *n_recsp is populated with the number of records returned.
 * If there are no pending records, NULL is returned and *n_recsp is set to 0.
 * If there are records and the buffer is not large enough to accomodate a
 * single entry, NULL is returned and *n_recsp is non-zero.
 * Only a string representation of the value is returned but the key and field
 * are in raw format.
 * All records obtained must be marked as completed before more can be
 * retrieved. A record is considered completed once kvstore_jrnl_records_act()
 * is called on it without the KVSTORE_JRNL_ACT_WILL_RETRY action in it.
 */
struct kvstore_jrnl_rec *kvstore_jrnl_get_records(struct kvstore_entry_id *id,
    void *buf, size_t bufsz, size_t *n_recsp);

/*
 * Indicate a set of records should be acted upon by the kvstore.
 * For each record the actions field is interpreted to modify corresponding
 * kvstore entries.
 * A record not having KVSTORE_JRNL_ACT_WILL_RETRY must not receive any
 * subsequent kvstore_jrnl_records_act() call and a record having
 * KVSTORE_JRNL_ACT_WILL_RETRY must receive a subsequent subsequent call
 * to mark the record as completed.
 * Records can be acted upon in any order compared to the order of
 * kvstore_jrnl_get_records() and in as many calls as necessary.
 */
void kvstore_jrnl_records_act(struct kvstore *kv,
    const struct kvstore_jrnl_rec *recs, size_t n_recs);
