/*-
 * Copyright (c) 2024 Varnish Software AS
 * All rights reserved.
 *
 * Author: Stephane Cance <stephane.cance@varnish-software.com>
 *
 */
#ifdef KVSTORE_LIB_H_INCLUDED
#  error "kvstore_lib.h included multiple times"
#endif /* KVSTORE_LIB_H_INCLUDED */
#define KVSTORE_LIB_H_INCLUDED

struct kvstore_entry_id {
	unsigned			    magic;
#define KVSTORE_ENTRY_ID_MAGIC		    0x65a9fb71
	uint32_t			    khash;
	uint64_t			    uid;
	struct kvstore			    *kv;
};

/* any field in this struct can be NULL */
struct kvstore_counters {
	struct VSC_lck			    *vsc_lck_ul;
	struct VSC_lck			    *vsc_lck_jrnl;
	struct VSC_cond			    *vsc_cond_jrnl;
	uint64_t			    *n_lck_rp;
	uint64_t			    *n_lck_up;
	uint64_t			    *n_lck_u2wp;
	uint64_t			    *n_lck_wp;
	uint64_t			    *n_jrnl_act_setp;
	uint64_t			    *n_jrnl_act_delp;
	uint64_t			    *n_jrnl_syncp;
	uint64_t			    *n_jrnl_trustedp;
	uint64_t			    *n_jrnl_keptp;
};

struct kvstore *kvstore_create(unsigned n_buckets, uint32_t hash_seed);
struct kvstore *kvstore_create_ex(unsigned n_buckets, uint32_t hash_seed,
    const struct kvstore_counters *cnt);
void kvstore_destroy(struct kvstore **kvp);

/* return the number of keys currently in the kvstore */
size_t kvstore_count(struct kvstore *kv);

/* return non-zero if a key is removed */
int kvstore_remove(struct kvstore *kv, const void *key, size_t klen);
int kvstore_compo_remove(struct kvstore *kv, const void *key, size_t klen,
    const void *field, size_t fieldlen);

/* returns non-zero if a non-expired key exists (of any type). */
int kvstore_key_exists(struct kvstore *kv, const void *key, size_t klen);

/*
 * Iterates over up to n entries from a saved up position and apply an
 * expiration check on the entries.
 * All values of khash and uid are valid and can be artbitrarily set to
 * start iterating at a random position.
 * *id is updated during the call to the new iterator position. The iterator
 * resets itself seemlessly to the begining once a full loop is done.
 * Returns the number of entries expired.
 */
size_t kvstore_expiration_check(struct kvstore_entry_id *id, size_t n);

/* same as above but allows to fake elapsed time for testing purposes */
size_t kvstore_expiration_check_at(struct kvstore_entry_id *id, unsigned n,
    vtim_mono now);

/*
 * int64 by key API
 * if id is not NULL, *id is populated with the id of the manipulated entry.
 */

/* Returns 0 if no valid value was found */
int kvstore_int64_get(struct kvstore *kv, const void *key, size_t klen,
    int64_t *vp, vtim_dur *ttlp);
void kvstore_int64_set(struct kvstore *kv, const void *key, size_t klen,
    int64_t v, vtim_dur ttl);

int64_t kvstore_int64_fetch_and_add(struct kvstore *kv, const void *key,
    size_t klen, int64_t v, vtim_dur ttl, struct kvstore_entry_id *id);
int64_t kvstore_int64_add_and_fetch(struct kvstore *kv, const void *key,
    size_t klen, int64_t v, vtim_dur ttl, struct kvstore_entry_id *id);

/*
 * int64 by id API
 * It is a programming error to provide an id pointing to an entry which is not
 * of int64 type (this is asserted).
 * Contrary to the API leveraging key and klen this API ensures that the
 * operation touches the exact same key as the operation that looked up the id.
 * Returns non-zero on success.
 */

int kvstore_id_int64_fetch_and_add(const struct kvstore_entry_id *id,
    int64_t v, int64_t *vp);
int kvstore_id_int64_add_and_fetch(const struct kvstore_entry_id *id,
    int64_t v, int64_t *vp);

/*
 * int64 composite field by key API
 * if compoid is not NULL, *compoid is populated with the id of the composite
 * entry indexing the field.
 * if id is not NULL, *id is populated with the id of the manipulated field of
 * the composite entry.
 */

/* Returns 0 if no valid value was found */
int kvstore_compo_int64_get(struct kvstore *kv, const void *key, size_t klen,
    const void *field, size_t fieldlen, int64_t *vp, vtim_dur *ttlp);
void kvstore_compo_int64_set(struct kvstore *kv, const void *key, size_t klen,
    const void *field, size_t fieldlen, int64_t v, vtim_dur ttl);

int64_t kvstore_compo_int64_fetch_and_add(struct kvstore *kv, const void *key,
    size_t klen, const void *field, size_t fieldlen, int64_t v, vtim_dur ttl,
    struct kvstore_entry_id *compoid, struct kvstore_entry_id *id);
int64_t kvstore_compo_int64_add_and_fetch(struct kvstore *kv, const void *key,
    size_t klen, const void *field, size_t fieldlen, int64_t v, vtim_dur ttl,
    struct kvstore_entry_id *compoid, struct kvstore_entry_id *id);

/*
 * int64 composite field by id API
 * It is a programming error to provide ids pointing to anything but a field of
 * a int64 type (this is asserted).
 * Contrary to the API leveraging key and klen this API ensures that the
 * operation touches the exact same key as the operation that looked up the ids.
 * Returns non-zero on success.
 */

int kvstore_compo_id_int64_fetch_and_add(const struct kvstore_entry_id *compoid,
    const struct kvstore_entry_id *id, int64_t v, int64_t *vp);
int kvstore_compo_id_int64_add_and_fetch(const struct kvstore_entry_id *compoid,
    const struct kvstore_entry_id *id, int64_t v, int64_t *vp);

/*
 * double by key API
 */

/* Returns 0 if no valid value was found */
int kvstore_dbl_get(struct kvstore *kv, const void *key, size_t klen,
    double *vp, vtim_dur *ttlp);
void kvstore_dbl_set(struct kvstore *kv, const void *key, size_t klen,
    double v, vtim_dur ttl);

/* Returns 0 if no valid value was found */
int kvstore_compo_dbl_get(struct kvstore *kv, const void *key, size_t klen,
    const void *field, size_t fieldlen, double *vp, vtim_dur *ttlp);
void kvstore_compo_dbl_set(struct kvstore *kv, const void *key, size_t klen,
    const void *field, size_t fieldlen, double v, vtim_dur ttl);

/*
 * string by key API
 */

/*
 * Automatically converts int64 to (NUL-terminated) string representation.
 * buf and bufsz follow snprintf() like semantics, e.g. providing a
 * NULL buf is only valid if bufsz is 0.
 * *vszp returns the length that was stored which may be larger than bufsz.
 * If try_conv is non-zero allow some type conversions to string.
 */
int kvstore_str_get(struct kvstore *kv, const void *key, size_t klen,
    char *buf, size_t bufsz, size_t *vszp, vtim_dur *ttlp, int try_conv);

/*
 * If the caller wants a terminating '\0' byte to be returned during
 * a subsequent kvstore_str_get() then val must contain one and vsz
 * include it.
 */
void kvstore_str_set(struct kvstore *kv, const void *key, size_t klen,
    const char *val, size_t vsz, vtim_dur ttl);

/*
 * Automatically converts int64 to (NUL-terminated) string representation.
 * buf and bufsz follow snprintf() like semantics, e.g. providing a
 * NULL buf is only valid if bufsz is 0.
 * *vszp returns the length that was stored which may be larger than bufsz.
 * If try_conv is non-zero allow some type conversions to string.
 */
int kvstore_compo_str_get(struct kvstore *kv, const void *key, size_t klen,
    const void *field, size_t fieldlen, char *buf, size_t bufsz, size_t *vszp,
    vtim_dur *ttlp, int try_conv);

/*
 * If the caller wants a terminating '\0' byte to be returned during
 * a subsequent kvstore_str_get() then val must contain one and vsz
 * include it.
 */
void kvstore_compo_str_set(struct kvstore *kv, const void *key, size_t klen,
    const void *field, size_t fieldlen, const char *val, size_t vsz,
    vtim_dur ttl);
